package ims.hotcoref.data;

import ims.hotcoref.data.CFGTree.CFGNode;
import ims.hotcoref.symbols.ISymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.IntPair;

import java.util.HashMap;
import java.util.Map;

public abstract class AbstractSentence implements ISentence {

	private static final String HYPHEN="-";
	private static final String BAR="|";
	
	protected final int sentenceIndex;
	protected final String[] corefCol;
	protected final Document d;
	protected SInst sInst;
	
	protected final int[] quoteCount;
	protected final String[] forms;
	
	private Map<IntPair,Span> spanMap=new HashMap<IntPair,Span>();
	
	protected AbstractSentence(int sentenceIdx,String[] corefCol,Document d,String[] forms){
		this.sentenceIndex=sentenceIdx;
		this.corefCol=corefCol;
		this.d=d;
		this.forms=forms;
		this.quoteCount=new int[forms.length];
	}

	@Override
	public void clearCorefCol(){
		for(int i=1;i<corefCol.length;++i){
			corefCol[i]=HYPHEN;
		}
	}

	public void addCoref(Span s, Integer key) {
		if(s.start==s.end){ //Single token
			String c="("+key+")";
//			if(corefCol[s.start].contains(c)) //XXX remove this wehn done debugging
//				System.out.println("HERE!");
			if(corefCol[s.start].equals(HYPHEN)){
				corefCol[s.start]=c;
			} else {
				corefCol[s.start]+=BAR+c;
			}
			
		} else { //Multiple tokens
			String b="("+key;
			String e=key+")";
//			if(corefCol[s.start].contains(b) && corefCol[s.end].contains(e)) //XXX remove this too
//				System.out.println("HERE!");
			if(corefCol[s.start].equals(HYPHEN)){
				corefCol[s.start]=b;
			} else {
				corefCol[s.start]+=BAR+b;
			}
			if(corefCol[s.end].equals(HYPHEN)){
				corefCol[s.end]=e;
			} else {
				corefCol[s.end]+=BAR+e;
			}
		}
	}
	
	@Override
	public String[] getCorefCol(){
		return corefCol;
	}
	
	@Override
	public Span getSpan(int beg, int end){
		//XXX
//		IntPair ip=IntPair.intPairPool.get(beg, end);
		CFGTree ct=getCFGTree();
		IntPair ip=new IntPair(beg,end);
		synchronized(spanMap){
			Span s=spanMap.get(ip);
			if(s==null){
				CFGNode cfgNode=(ct==null?null:ct.getExactNode(beg,end));
				s=new Span(this,beg,end,cfgNode);
				spanMap.put(ip,s);
				return s;
			} else {
				return s;
			}
		}
	}
	
	@Override
	public Map<IntPair,Span> getSpanMap(){
		return spanMap;
	}

	@Override
	public int getSentenceIndex(){
		return sentenceIndex;
	}
	
	public void initInst(SymbolTable symTab){
		int[] f  = symTab.forms!=null   ? s2iArr(getForms(),symTab.forms)      : null;
		int[] l  = symTab.lemmas!=null  ? s2iArr(getLemmas(),symTab.lemmas)    : null;
		int[] t  = symTab.tags!=null    ? s2iArr(getTags(),symTab.tags)        : null;
		int[] b  = symTab.bwuv!=null    ? s2iArr(getBWUV(),symTab.bwuv)        : null;
		int[] d  = symTab.deprels!=null ? s2iArr(getDepTree().lbls,symTab.deprels)  : null;
		int[] fc = symTab.chars!=null   ? getFirstCharArr(getForms(),symTab.chars) : null;
		int[] lc = symTab.chars!=null   ? getLastCharArr(getForms(),symTab.chars)  : null;
		int[] f2c= symTab.charBigrams!=null ? getFirstCharBG(getForms(),symTab.charBigrams) : null;
		int[] l2c= symTab.charBigrams!=null ? getLastCharBG(getForms(),symTab.charBigrams)  : null;
		int[] accent= symTab.accent!=null ? s2iArr(getAccent(),symTab.accent)  : null;
		int[] boundary= symTab.boundary!=null ? s2iArr(getBoundary(),symTab.boundary)  : null;
		int[] nuclear= symTab.nuclear!=null ? s2iArr(getNuclear(),symTab.nuclear)  : null;
		int[] tobitype= symTab.tobitype!=null ? s2iArr(getTobiType(),symTab.tobitype)  : null;

		sInst=new SInst(f,l,t,b,d,fc,lc,f2c,l2c,accent,boundary,nuclear,tobitype);
	}

	private int[] getFirstCharBG(String[] forms,ISymbolMapping<String> charBigrams) {
		int[] i=new int[forms.length];
		for(int iq=0;iq<forms.length;++iq)
			i[iq]=charBigrams.lookup(forms[iq].substring(0,2));
		return i;
	}

	private int[] getLastCharBG(String[] forms,ISymbolMapping<String> charBigrams) {
		int[] i=new int[forms.length];
		for(int iq=0;iq<forms.length;++iq){
			int l=forms[iq].length();
			i[iq]=charBigrams.lookup(forms[iq].substring(l-2,l));
		}
		return i;
	}

	private int[] getLastCharArr(String[] forms, ISymbolMapping<String> chars) {
		int[] i=new int[forms.length];
		for(int iq=0;iq<forms.length;++iq){
			int l=forms[iq].length();
			i[iq]=chars.lookup(forms[iq].substring(l-1,l));
		}
		return i;
	}

	private int[] getFirstCharArr(String[] forms, ISymbolMapping<String> chars) {
		int[] i=new int[forms.length];
		for(int iq=0;iq<forms.length;++iq)
			i[iq]=chars.lookup(forms[iq].substring(0,1));
		return i;
	}

	private int[] s2iArr(String[] s, ISymbolMapping<String> ssm) {
		int[] i=new int[s.length];
		for(int iq=0;iq<s.length;++iq)
			i[iq]=ssm.lookup(s[iq]);
		return i;
	}


	@Override
	public int getLength() {
		return forms.length;
	}
	
	@Override
	public String[] getForms() {
		return forms;
	}

	@Override
	public String getForm(int i) {
		return forms[i];
	}
	
	
	@Override
	public String[] getAccent() {
		return forms;
	}

	@Override
	public String getAccent(int i) {
		return forms[i];
	}
	
	@Override
	public String[] getBoundary() {
		return forms;
	}

	@Override
	public String getBoundary(int i) {
		return forms[i];
	}
	
	@Override
	public String[] getNuclear() {
		return forms;
	}

	@Override
	public String getNuclear(int i) {
		return forms[i];
	}

	@Override
	public String[] getTobiType() {
		return forms;
	}

	@Override
	public String getTobiType(int i) {
		return forms[i];
	}


	@Override
	public void setForm(int i, String string) {
		forms[i]=string;
	}

	
	@Override
	public SInst getSInst() {
		return sInst;
	}
	
	public int[] getQuoteCount(){
		return quoteCount;
	}


}
