package ims.hotcoref.data;

import ims.hotcoref.io.ColumnReader;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class DIRNDLAnaphSentence extends AbstractSentence {


		static final int FORM_COL=6; //3
		static final int TAG_COL=4;
		static final int CFG_COL=5;
		public static final int LEMMA_COL=3; //6
		static final int SPEAKER_COL=9;
		static final int ToBI_COL=9;
		static final int NE_COL=10; 
		static final int GEND_COL=8;
		static final int NUM_COL=7;
		static final int TOBITYPE_COL=11;
		static final int BOUNDARY_COL=12;
		static final int NUCLEAR_COL=13;


		
		protected final String[] forms;
		protected final String[] lemmas;
		protected final String[] tags;
		protected final String[] cfgCol;
		protected final String[] neCol;
		protected final CFGTree ct;
		protected final List<NE> nes;
		protected final String[] speaker;
		protected final String[] originalForm;
		protected final String[] originalTags;
		protected final String[] originalAccent;
		protected final String[] bwuv;	
		protected final String[] num;
		protected final String[] gend;
		protected final String[] accent;
		protected final String[] tobitype;
		protected final String[] boundary;
		protected final String[] nuclear;


		
//		private Collection<TypedDependency> stanfordDeps;
		

		
		public static DIRNDLAnaphSentence makeSentence(List<String> lines,Integer sentenceIdx,Document d){
			String[] forms=new String[lines.size()+1];
			String[] lemmas=new String[lines.size()+1];
			String[] numCol=new String[lines.size()+1];
			String[] gendCol=new String[lines.size()+1];
			String[] tags=new String[lines.size()+1];
			String[] cfgCol=new String[lines.size()+1];
			String[] neCol=new String[lines.size()+1];
			String[] speakerCol=new String[lines.size()+1];
			String[] corefCol=new String[lines.size()+1];
			String[] accentCol=new String[lines.size()+1];
			String[] tobiTypeCol=new String[lines.size()+1];
			String[] boundaryCol=new String[lines.size()+1];
			String[] nuclearCol=new String[lines.size()+1];

			
			forms[0]=ROOT_FORM;
			numCol[0]=ROOT_NUM;
			gendCol[0]=ROOT_GEND;
			lemmas[0]=ROOT_LEMMA;
			tags[0]=ROOT_TAG;
			cfgCol[0]=ROOT_CFG;
			neCol[0]=ROOT_NE;
			accentCol[0]=ROOT_ACCENT;
			tobiTypeCol[0]=ROOT_ACCENT;
			boundaryCol[0]=ROOT_ACCENT;
			nuclearCol[0]=ROOT_ACCENT;
			speakerCol[0]=ROOT_SPEAKER;
			corefCol[0]=ROOT_COREF;
			
			for(int i=0;i<lines.size();++i){
				String line=lines.get(i);
				String[] cols=ColumnReader.WS.split(line);
				forms[i+1]=cols[FORM_COL];
				lemmas[i+1]=cols[LEMMA_COL];
				tags[i+1]=cols[TAG_COL];
				cfgCol[i+1]=cols[CFG_COL];
				neCol[i+1]=cols[NE_COL];
				speakerCol[i+1]=cols[SPEAKER_COL];
				accentCol[i+1]=cols[ToBI_COL];
				tobiTypeCol[i+1]=cols[TOBITYPE_COL];
				boundaryCol[i+1]=cols[BOUNDARY_COL];
				nuclearCol[i+1]=cols[NUCLEAR_COL];
				corefCol[i+1]=cols[cols.length-1];
				numCol[i+1]=cols[NUM_COL];
				gendCol[i+1]=cols[GEND_COL];

			}
			//		
			DIRNDLAnaphSentence s=new DIRNDLAnaphSentence(sentenceIdx,forms,tags,corefCol,speakerCol,neCol,cfgCol,numCol,gendCol,accentCol,boundaryCol,nuclearCol,tobiTypeCol,lemmas,d);

			return s;
		}
		
		@Override
		public List<String[]> getLineColumns() {
			List<String[]> r=new ArrayList<String[]>();
			
			String[] docName=new String[getLength()];
			Arrays.fill(docName, d.docName);
			r.add(docName);
			
			String[] docNo=new String[getLength()];
			Arrays.fill(docNo, d.docNo);
			r.add(docNo);
			
			String[] tokIdx=new String[getLength()];
			for(int i=0;i<tokIdx.length;++i) { tokIdx[i]=Integer.toString(i); }
			r.add(tokIdx);
			
			r.add(lemmas);
			r.add(originalTags);
		//	r.add(originalAccent);
			r.add(cfgCol);
			r.add(originalForm);
			r.add(num);
			r.add(gend);
			r.add(originalAccent);
			r.add(tobitype);
			r.add(nuclear);
			r.add(boundary);
			
			String[] senseSrlStuff=new String[getLength()];
			Arrays.fill(senseSrlStuff, "-");
			
			//r.add(senseSrlStuff);
		//	r.add(senseSrlStuff);
			//r.add(senseSrlStuff);
			
			//r.add(speaker);
			r.add(neCol);
			
			r.add(corefCol);
			
			return r;
		}
		
		public DIRNDLAnaphSentence(int index,String[] forms,String[] tags,String[] corefs,String[] speaker,String[] neCol,String[] cfgCol,String[] numCol, String[] gendCol,String[] accentCol,String[] boundaryCol,String[] nuclearCol, String[] tobiTypeCol,String[] lemmas, Document d){
			super(index,corefs,d,forms);
			this.forms=forms;
			this.lemmas=lemmas;
			this.tags=tags;
			this.gend=gendCol;
			this.accent=accentCol;
			this.boundary=boundaryCol;
			this.tobitype=tobiTypeCol;
			this.nuclear=nuclearCol;
			this.num=numCol;
			this.speaker=speaker;
			this.neCol=neCol;
			this.cfgCol=cfgCol;
			this.bwuv=new String[forms.length];
			ct=new CFGTree(cfgCol,this);
			nes=NE.getNEs(neCol, this);
			this.originalForm=Arrays.copyOf(forms, forms.length);
			this.originalTags=Arrays.copyOf(tags, tags.length);
			this.originalAccent=Arrays.copyOf(accent, accent.length);

		}

		public String toString(){
			StringBuilder sb=new StringBuilder();
			for(int i=0;i<forms.length;++i){
				sb.append(forms[i]).append(" ");
			}
			return sb.toString();
		}	
		
		@Override
		public String getLemma(int i) {
			return lemmas[i];
		}

		@Override
		public String getTag(int i) {
			return tags[i];
		}

		@Override
		public String[] getTags() {
			return tags;
		}
		
		@Override
		public String[] getAccent() {
			return accent;
		}
		
		@Override
		public String getAccent(int i) {
			return accent[i];
		}
		
		@Override
		public String[] getTobiType() {
			return tobitype;
		}
		
		@Override
		public String getTobiType(int i) {
			return tobitype[i];
		}
		
		
		@Override
		public String[] getNuclear() {
			return nuclear;
		}
		
		@Override
		public String getNuclear(int i) {
			return nuclear[i];
		}
		@Override
		public String[] getBoundary() {
			return boundary;
		}
		
		@Override
		public String getBoundary(int i) {
			return boundary[i];
		}
		
		
		@Override
		public String getNum(int i) {
			return num[i];
		}
		@Override
		public String[] getNum() {
			return num;
		}
		@Override
		public String getGend(int i) {
			return gend[i];
		}
		@Override
		public String[] getGend() {
			return gend;
		}


		@Override
		public void setTag(int i, String string) {
			tags[i]=string;
		}

		@Override
		public CFGTree getCFGTree() {
			return ct;
		}
		
		@Override
		public String getSpeaker(int i) {
			return speaker[i];
		}

		@Override
		public DepTree getDepTree() {
			return null;
		}

		@Override
		public String getBWUV(int i) {
			return bwuv[i];
		}

		@Override
		public void setBWUV(int i, String s) {
			bwuv[i]=s;
		}

		@Override
		public List<NE> getNEs() {
			return nes;
		}

		@Override
		public String[] getLemmas() {
			return lemmas;
		}

		@Override
		public String[] getBWUV() {
			return bwuv;
		}

		@Override
		public void setNum(int i, String string) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void setGend(int i, String string) {
			// TODO Auto-generated method stub
			
		}



		@Override
		public void setAccent(int i, String string) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void setBoundary(int i, String string) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void setNuclear(int i, String string) {
			// TODO Auto-generated method stub
			
		}

		@Override
		public void setTobiType(int i, String string) {
			// TODO Auto-generated method stub
			
		}

//		public Collection<TypedDependency> getStanfordDeps() {
//			StringBuilder sb=new StringBuilder();
//			for(int i=1;i<forms.length;++i)
//				sb.append(cfgCol[i].replace("*","("+tags[i]+" "+forms[i]+")"));
//			if(stanfordDeps==null)
//				try {
//					stanfordDeps=StanfordDepConverter.CONVERTER_SINGLETON.getBasicDependencies(sb.toString());
//				} catch(IllegalArgumentException e){ //Happens for e.g. NOPARSE
//					return null;
//				}
//			return stanfordDeps;
//		}

	}