package ims.hotcoref.data;

import ims.hotcoref.symbols.SymbolTable;
import ims.util.IntPair;

import java.util.List;
import java.util.Map;

public interface ISentence {

	static final String ROOT_FORM="<root-form>";
	static final String ROOT_LEMMA="<root-lemma>";
	static final String ROOT_TAG="<root-tag>";
	static final String ROOT_CFG="<root-cfg-col>";
	static final String ROOT_NE="<root-ne>";
	static final String ROOT_SPEAKER="<root-speaker>";
	static final String ROOT_COREF="<coref-null>";
	static final String ROOT_HEAD="-1";
	static final String ROOT_LBL="<root-lbl>";
	static final String[] ROOT_FEATS={"<root-feat>"};
	static final String ROOT_NUM="<root-num>";
	static final String ROOT_GEND="<root-gend>";
	static final String ROOT_ACCENT="<root-accent>";

	
	public static final String TAB_REGEXP = "\t";
	
	
	//The first List is a list of columns, the second by line. SO,
	//getLineColumns().get(3)[5] should correspond to the 4th column and the 5th line.
	public List<String[]> getLineColumns();
	
	
	/**
	 * get sentence length (including root token)
	 * @return length of sentence
	 */
	public int getLength();
	
	/**
	 * 
	 * @return index of sentence in document
	 */
	public int getSentenceIndex();
	
	public SInst getSInst();
	
	//Form
	public String[] getForms();
	public String getForm(int i);
	public void setForm(int i, String string);
	
	//Lemma
	public String[] getLemmas();
	public String getLemma(int i);
	
	//BWUV
	public String[] getBWUV();
	public String getBWUV(int i);
	public void setBWUV(int i,String s);
	
	//Tag
	public String[] getTags();	
	public String getTag(int i);
	public void setTag(int i, String string);
	
	//Num
	public String[] getNum();
	public String getNum(int i);
	public void setNum(int i, String string);

	//Gend

	public String[] getGend();
	public String getGend(int i);
	public void setGend(int i, String string);
	
	//Accent
	public String[] getAccent();
	public String getAccent(int i);
	public void setAccent(int i, String string);

	//CFG tree
	public CFGTree getCFGTree();
	
	//Dep tree
	public DepTree getDepTree();

	public List<NE> getNEs();
	
	//Speaker
	public String getSpeaker(int i);


	public void clearCorefCol();
	public String[] getCorefCol();
	public void addCoref(Span s, Integer chainId);


	public Span getSpan(int beg, int end);
	public Map<IntPair, Span> getSpanMap();
	
	public int[] getQuoteCount();


	public void initInst(SymbolTable symbolTable);

	
	public String[] getBoundary();
	public String getBoundary(int i);
	public void setBoundary(int i, String string);


	public String[] getNuclear();
	public String getNuclear(int i);
	public void setNuclear(int i, String string);
	
	public String[] getTobiType();
	public String getTobiType(int i);
	public void setTobiType(int i, String string);


}
