package ims.hotcoref.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


public class SemEval10Sentence extends AbstractSentence {

	
	
	static final int FORM_IDX=1;
	static final int LEMMA_IDX=2;
	static final int PLEMMA_IDX=3;
	static final int POS_IDX=4;
	static final int PPOS_IDX=5;
	static final int FEAT_IDX=6;
	static final int PFEAT_IDX=7;
	static final int HEAD_IDX=8;
	static final int PHEAD_IDX=9;
	static final int DEPREL_IDX=10;
	static final int PDEPREL_IDX=11;
	static final int NE_IDX=12;
	static final int PNE_IDX=13;

	protected final String[] lemma;
	protected final String[] plemma;
	protected final String[] pos;
	protected final String[] ppos;
	protected final String[] feat;
	protected final String[] pfeat;
	protected final String[] head;
	protected final String[] phead;
	protected final String[] deprel;
	protected final String[] pdeprel;
	protected final String[] ne;
	protected final String[] pne;
	
	protected final DepTree dt;
	protected final List<NE> nes;
	
	protected SemEval10Sentence(int sentenceIdx, Document d, 
			String[] forms,String[] lemma,String[] plemma,String[] pos,String[] ppos,String[] feat,String[] pfeat,
			String[] head, String[] phead,String[] deprel,String[] pdeprel,String[] ne, String[] pne,
			String[] corefCol) {
		super(sentenceIdx, corefCol, d, forms);
		this.lemma=lemma;
		this.plemma=plemma;
		this.pos=pos;
		this.ppos=ppos;
		this.feat=feat;
		this.pfeat=pfeat;
		this.head=head;
		this.phead=phead;
		this.deprel=deprel;
		this.pdeprel=pdeprel;
		this.ne=ne;
		this.pne=pne;
		
		this.dt=new DepTree(phead,pdeprel);
		this.nes=NE.getNEs(pne, this);
	}
	
	public static SemEval10Sentence fromLines(List<String> lines,int sentIdx,Document d){
		String[] forms=new String[lines.size()+1];
		String[] lemma=new String[lines.size()+1];
		String[] plemma=new String[lines.size()+1];
		String[] pos=new String[lines.size()+1];
		String[] ppos=new String[lines.size()+1];
		String[] feat=new String[lines.size()+1];
		String[] pfeat=new String[lines.size()+1];
		String[] head=new String[lines.size()+1];
		String[] phead=new String[lines.size()+1];
		String[] deprel=new String[lines.size()+1];
		String[] pdeprel=new String[lines.size()+1];
		String[] ne=new String[lines.size()+1];
		String[] pne=new String[lines.size()+1];
		String[] corefCol=new String[lines.size()+1];
		forms[0]=ROOT_FORM;
		lemma[0]=ROOT_LEMMA;
		plemma[0]=ROOT_LEMMA;
		pos[0]=ROOT_TAG;
		ppos[0]=ROOT_TAG;
		feat[0]=ROOT_FEATS[0];
		pfeat[0]=ROOT_FEATS[0];
		head[0]=ROOT_HEAD;
		phead[0]=ROOT_HEAD;
		deprel[0]=ROOT_LBL;
		pdeprel[0]=ROOT_LBL;
		ne[0]=ROOT_NE;
		pne[0]=ROOT_NE;
		corefCol[0]=ROOT_COREF;
		int i=1;
		for(String line:lines){
			String[] a=line.split(TAB_REGEXP);
			forms[i]=a[FORM_IDX];
			lemma[i]=a[LEMMA_IDX];
			plemma[i]=a[PLEMMA_IDX];
			pos[i]=a[POS_IDX];
			ppos[i]=a[PPOS_IDX];
			feat[i]=a[FEAT_IDX];
			pfeat[i]=a[PFEAT_IDX];
			head[i]=a[HEAD_IDX];
			phead[i]=a[PHEAD_IDX];
			deprel[i]=a[DEPREL_IDX];
			pdeprel[i]=a[PDEPREL_IDX];
			ne[i]=a[NE_IDX];
			pne[i]=a[PNE_IDX];
			corefCol[i]=a[a.length-1];			
			++i;
		}
		return new SemEval10Sentence(sentIdx, d, forms, lemma, plemma, pos, ppos, feat, pfeat, head, phead, deprel, pdeprel, ne, pne, corefCol);
	}

	@Override
	public List<String[]> getLineColumns() {
		List<String[]> r=new ArrayList<String[]>(); 

		String[] id=new String[forms.length];
		for(int i=0;i<id.length;++i)
			id[i]=Integer.toString(i);
		
		
		r.add(id);
		r.add(forms);
		r.add(lemma);
		r.add(plemma);
		r.add(pos);
		r.add(ppos);
		r.add(feat);
		r.add(pfeat);
		r.add(head);
		r.add(phead);
		r.add(deprel);
		r.add(pdeprel);
		r.add(ne);
		r.add(pne);
		
		String[] US=new String[forms.length];
		Arrays.fill(US, "_");
		
		r.add(US);
		r.add(US);
		
		r.add(corefCol);
		
		return r;
	}

	@Override
	public String[] getLemmas() {
		return plemma;
	}

	@Override
	public String getLemma(int i) {
		return plemma[i];
	}

	@Override
	public String[] getBWUV() {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public String getBWUV(int i) {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public void setBWUV(int i, String s) {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public String[] getTags() {
		return ppos;
	}
	
	@Override
	public String getTag(int i) {
		return ppos[i];
	}
	
	@Override
	public String[] getGend() {
		throw new UnsupportedOperationException("not supported for this format");
	}
	
	public String[] getAccent() {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public String getGend(int i) {
		throw new UnsupportedOperationException("not supported for this format");
	}
	
	@Override
	public String[] getNum() {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public String getNum(int i) {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public void setTag(int i, String string) {
		ppos[i]=string;
	}

	@Override
	public CFGTree getCFGTree() {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public DepTree getDepTree() {
		return dt;
	}

	@Override
	public List<NE> getNEs() {
		return nes;
	}

	@Override
	public String getSpeaker(int i) {
		throw new UnsupportedOperationException("not supported for this format");
	}

	@Override
	public void setNum(int i, String string) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void setGend(int i, String string) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public String getAccent(int i) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void setAccent(int i, String string) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void setBoundary(int i, String string) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void setNuclear(int i, String string) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void setTobiType(int i, String string) {
		// TODO Auto-generated method stub
		
	}


}
