package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.hotcoref.util.CompoundSplitLookup;
import ims.util.ThreadPoolSingleton;

public class F_CompoundHeadMatch extends AbstractPairEnumFeature<TrueFalse>{
	private static final long serialVersionUID = 1736992482622613222L;

	private final PairTargetNodeExtractor span1;
	private final PairTargetNodeExtractor span2;
	
	
	protected F_CompoundHeadMatch(PairTargetNodeExtractor span1Node,PairTargetNodeExtractor span2Node) {
		super(span1Node.ts.toString()+span2Node.ts.toString()+"CompoundHeadMatch", TrueFalse.values());
		this.span1=span1Node;
		this.span2=span2Node;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab,List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.compoundMatch!=null)
			return;
		inst.compoundMatch=new byte[2][][];
		inst.compoundMatch[0]=new byte[inst.nodes.length][]; //0 means the bigger span is to the left of the head
		inst.compoundMatch[1]=new byte[inst.nodes.length][]; //1 means the bigger span is to the right of the head
		
		Callable<Void> j=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int nT=1;nT<inst.nodes.length;++nT){
					inst.compoundMatch[0][nT]=new byte[nT];
					for(int nF=0;nF<nT;++nF)
						inst.compoundMatch[0][nT][nF]=getByte(nF,nT,inst);
				}
				return null;
			}
		};
		Callable<Void> j2=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int nT=1;nT<inst.nodes.length;++nT){
					inst.compoundMatch[1][nT]=new byte[nT];
					for(int nF=0;nF<nT;++nF)
						inst.compoundMatch[1][nT][nF]=getByte(nT,nF,inst);
				}
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(j));
		fjs.add(ThreadPoolSingleton.getInstance().submit(j2));
	}

	private byte getByte(int subIdx,int wholeIdx, Instance inst) {
//		INode sub=inst.nodes[subStr.getNodeInstIdx(nF,nT,null,inst)];
//		INode sup=inst.nodes[subStr.getOtherNodeInstIdx(nF,nT)];
//		INode sup=inst.nodes[wholeSpan.getNodeInstIdx(nF, nT, null, inst)];
		INode sub=inst.nodes[subIdx];
		INode sup=inst.nodes[wholeIdx];
		final TrueFalse t;
		if(sub.isVirtualNode() || sup.isVirtualNode())
			t=TrueFalse.VNode;
		else 
			t=getT((MNode) sub,(MNode) sup);
		return (byte) t.ordinal();
	}

	private TrueFalse getT(MNode sub, MNode sup) {
		
		
		String hdFormAnt=sub.span.s.getForm(sub.span.hd); 
		String hdFormAna=sup.span.s.getForm(sup.span.hd);
				
		if (CompoundSplitLookup.compounds.containsKey(hdFormAnt)){
			String value=CompoundSplitLookup.compounds.get(hdFormAnt);
			
			//antecedent's head matches anaphor
			if (value.equalsIgnoreCase(hdFormAna))		
				return TrueFalse.True;
			
			//both are compounds, head match
			if (CompoundSplitLookup.compounds.containsKey(hdFormAna)){
				String value2=CompoundSplitLookup.compounds.get(hdFormAna);
				
				if (value2.equalsIgnoreCase(value))
					return TrueFalse.True;
			}
	
		}
		
		//other way around
		else if (CompoundSplitLookup.compounds.containsKey(hdFormAna)){
			String value=CompoundSplitLookup.compounds.get(hdFormAna);
			
			//anaphor's head matches antecedent
			if (value.equalsIgnoreCase(hdFormAnt))		
				return TrueFalse.True;
				
		}
		
		return TrueFalse.False;

	}

	@Override
//	int getIntValue(Instance inst,int headIdx,int depIdx,HOTState hotState) {
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
//		int s=subStr.getNodeInstIdx(headIdx, depIdx, hotState);
//		int t=wholeSpan.getNodeInstIdx(headIdx, depIdx, hotState);
		int s=span1.getNodeInstIdxPrecompArray(tnes);
		int t=span2.getNodeInstIdxPrecompArray(tnes);

		final int l,r,i;
		if(s<t){
			l=s;
			r=t;
			i=1;
		} else {
			l=t;
			r=s;
			i=0;
		}
		if(l<0)
			return TrueFalse.None.ordinal();
		return inst.compoundMatch[i][r][l];
	}

	@Override
	public boolean firstOrderFeature() {
		return span1.firstOrder() && span2.firstOrder();
	}

}
