
package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.hotcoref.util.AnimacityLookup;
import ims.util.ThreadPoolSingleton;

public class F_GenderMatchNames extends AbstractPairEnumFeature<TrueFalse>{
	private static final long serialVersionUID = 1736992482622613222L;

	private final PairTargetNodeExtractor span1;
	private final PairTargetNodeExtractor span2;
	
	
	protected F_GenderMatchNames(PairTargetNodeExtractor span1Node,PairTargetNodeExtractor span2Node) {
		super(span1Node.ts.toString()+span2Node.ts.toString()+"GenderMatchNames", TrueFalse.values());
		this.span1=span1Node;
		this.span2=span2Node;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab,List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.nameMatch!=null)
			return;
		inst.nameMatch=new byte[2][][];
		inst.nameMatch[0]=new byte[inst.nodes.length][]; //0 means the bigger span is to the left of the head
		inst.nameMatch[1]=new byte[inst.nodes.length][]; //1 means the bigger span is to the right of the head
		
		Callable<Void> j=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int nT=1;nT<inst.nodes.length;++nT){
					inst.nameMatch[0][nT]=new byte[nT];
					for(int nF=0;nF<nT;++nF)
						inst.nameMatch[0][nT][nF]=getByte(nF,nT,inst);
				}
				return null;
			}
		};
		Callable<Void> j2=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int nT=1;nT<inst.nodes.length;++nT){
					inst.nameMatch[1][nT]=new byte[nT];
					for(int nF=0;nF<nT;++nF)
						inst.nameMatch[1][nT][nF]=getByte(nT,nF,inst);
				}
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(j));
		fjs.add(ThreadPoolSingleton.getInstance().submit(j2));
	}

	private byte getByte(int subIdx,int wholeIdx, Instance inst) {
//		INode sub=inst.nodes[subStr.getNodeInstIdx(nF,nT,null,inst)];
//		INode sup=inst.nodes[subStr.getOtherNodeInstIdx(nF,nT)];
//		INode sup=inst.nodes[wholeSpan.getNodeInstIdx(nF, nT, null, inst)];
		INode sub=inst.nodes[subIdx];
		INode sup=inst.nodes[wholeIdx];
		final TrueFalse t;
		if(sub.isVirtualNode() || sup.isVirtualNode())
			t=TrueFalse.VNode;
		else 
			t=getT((MNode) sub,(MNode) sup);
		return (byte) t.ordinal();
	}

	private TrueFalse getT(MNode sub, MNode sup) {
		
		Span ana=sup.span;
		Span ant=sub.span;
		
		String hdFormAna=sub.span.s.getLemma(sub.span.hd); //antecedent
		String hdFormAnt=sup.span.s.getLemma(sup.span.hd); //anaphor
		
		//anaphor fem or masc and antecedent pronoun of same gender
		//female
		if (sup.span.isPronoun &&  sup.span.number.name().equalsIgnoreCase("sin")&& sup.span.gender.name().equalsIgnoreCase("fem")){
			
			for(int r=ant.start;r<=ant.end;++r){
				if(AnimacityLookup.femaleList.contains(ant.s.getLemma(r)))
					return TrueFalse.True;
			}
		}
		
		//male
		else if (sup.span.isPronoun && sup.span.number.name().equalsIgnoreCase("sin") &&sup.span.gender.name().equalsIgnoreCase("masc")){
			
			for(int r=ant.start;r<=ant.end;++r){
				if(AnimacityLookup.maleList.contains(ant.s.getLemma(r))&&! ant.s.getLemma(r).equals("Mark"))
					return TrueFalse.True;
			}
		}
		
		/*//both are pronouns
		//check whether we should include this here or not! XXX
		 * for the time being: leave out
		else if (sup.span.isPronoun&& sub.span.isPronoun)
			return TrueFalse.True;*/
		
		return TrueFalse.False;
	}

	@Override
//	int getIntValue(Instance inst,int headIdx,int depIdx,HOTState hotState) {
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
//		int s=subStr.getNodeInstIdx(headIdx, depIdx, hotState);
//		int t=wholeSpan.getNodeInstIdx(headIdx, depIdx, hotState);
		int s=span1.getNodeInstIdxPrecompArray(tnes);
		int t=span2.getNodeInstIdxPrecompArray(tnes);

		final int l,r,i;
		if(s<t){
			l=s;
			r=t;
			i=1;
		} else {
			l=t;
			r=s;
			i=0;
		}
		if(l<0)
			return TrueFalse.None.ordinal();
		return inst.nameMatch[i][r][l];
	}

	@Override
	public boolean firstOrderFeature() {
		return span1.firstOrder() && span2.firstOrder();
	}

}
