package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_IsPronoun extends AbstractPairEnumFeature<TrueFalse>{
	
	private static final long serialVersionUID = 7460524611469166156L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_IsPronoun(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"IsPronoun", TrueFalse.values());
		this.nodeExtractor=nodeExtractor;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab,List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.isPronoun!=null)
			return;
		Callable<Void> c=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				inst.isPronoun=new byte[inst.nodes.length];
				for(int i=0;i<inst.nodes.length;++i)
					inst.isPronoun[i]=isPronoun(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(c));
	}

	@Override
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.isPronoun[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}

	private static byte isPronoun(INode iNode) {
		final TrueFalse t;
		if(iNode.isVirtualNode())
			t=TrueFalse.VNode;
		else {
			MNode mn=(MNode) iNode;
			if(mn.span.isPronoun)
				t=TrueFalse.True;
			else
				t=TrueFalse.False;
		}
		return (byte) t.ordinal();
	}

	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}
}