
package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.NuclearSequences;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_NuclearBagOfAccents extends AbstractPairEnumFeature<NuclearSequences> {
	private static final long serialVersionUID = 4988768071632378832L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_NuclearBagOfAccents(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"NuclearBagOfAccents", NuclearSequences.values());
		this.nodeExtractor=nodeExtractor;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab, List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.nuclearBagOfWords!=null)
			return;
		inst.nuclearBagOfWords=new byte[inst.nodes.length];
		Callable<Void> job=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int i=0;i<inst.nodes.length;++i)
					inst.nuclearBagOfWords[i]=getByte(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(job));
	}

	private byte getByte(INode iNode) {
		NuclearSequences n = null;
		if(iNode.isVirtualNode())
			n=NuclearSequences.none;
		else {
			MNode mn=(MNode) iNode;
			Span sp=mn.span;
			int spansize=mn.span.size();
			Boolean pn=false;
			Boolean n2=false;
			Boolean n1=false;
			
			for(int r=sp.start;r<=sp.end;++r)
				
				if (sp.s.getNuclear(r).equals("0")){}
				
				else if ((sp.s.getNuclear(r).equals("pn"))){//&&(spansize<=5)){					//&&(spansize<=3)   //n=NuclearSequences.n2;
					pn=true;
				}

				else if ((sp.s.getNuclear(r).equals("n1"))){//&&(spansize<=5)){
					n1=true;									//n1!
				}
			
				else if ((sp.s.getNuclear(r).equals("n2"))){//&&(spansize<=5)){
					n2=true;
				}	
			
			if (pn&n1&n2){
				n=NuclearSequences.pnn1n2;
			}
			else if (pn&n1&!n2){
				n=NuclearSequences.pnn1non2;
			}
			else if (pn&!n1&n2){
				n=NuclearSequences.pnnon1n2;
			}
			else if (pn&!n1&!n2){
				n=NuclearSequences.pnnon1non2;
			}
			else if (!pn&n1&n2){
				n=NuclearSequences.nopnn1n2; 
			}
			else if (!pn&n1&!n2){
				n=NuclearSequences.nopnn1non2;
			}
			else if (!pn&!n1&n2){
				n=NuclearSequences.nopnnon1non2;
			}
			else if (!pn&!n1&!n2){
				n=NuclearSequences.nopnnon1non2;
			}
			
			else {
				n=NuclearSequences.none;
			}
			

		}
		
		return (byte) n.ordinal();
	}

	@Override
//	int getIntValue(Instance inst,int headIdx,int depIdx,HOTState hotState) {
//		return inst.nuclear[nodeExtractor.getNodeInstIdx(headIdx,depIdx,hotState)];
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.nuclearBagOfWords[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}
	
	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}

}