
package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_NuclearFeature extends AbstractPairEnumFeature<TrueFalse>{
	
	private static final long serialVersionUID = 7460524611469166156L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_NuclearFeature(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"NuclearAccent", TrueFalse.values());
		this.nodeExtractor=nodeExtractor;
	}



	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}


	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab,List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.NuclearAccent!=null)
			return;
		Callable<Void> c=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				inst.NuclearAccent=new byte[inst.nodes.length];
				for(int i=0;i<inst.nodes.length;++i)
					inst.NuclearAccent[i]=NuclearAccent(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(c));
	}



	private static byte NuclearAccent(INode iNode) {
		 TrueFalse t;
		if(iNode.isVirtualNode())
			t=TrueFalse.VNode;
		else {
			MNode mn=(MNode) iNode;
			Span sp=mn.span;
			int spansize=mn.span.size();

			//String hdAccent=mn.span.s.getAccent(mn.span.hd); 
			
			for(int r=sp.start;r<=sp.end;++r)				
					//if((sp.s.getNuclear(r).equals("n1")||sp.s.getNuclear(r).equals("n2"))&&(spansize<=5))
					//if((sp.s.getNuclear(r).equals("n1")||(sp.s.getNuclear(r).equals("pn")||sp.s.getNuclear(r).equals("n2"))))//&&(spansize<=5))
				 if(sp.s.getNuclear(r).equals("n2") )//&(spansize<=5))
						return (byte) TrueFalse.True.ordinal();
			
			
		}
		return (byte) TrueFalse.False.ordinal();
	}


//original version of this workspace version	
/*private static byte NuclearAccent(INode iNode) {
	 TrueFalse t;
	if(iNode.isVirtualNode())
		t=TrueFalse.VNode;
	else {
		MNode mn=(MNode) iNode;
		Span sp=mn.span;
		int spansize=mn.span.size();

		//String hdAccent=mn.span.s.getAccent(mn.span.hd); 
		
		for(int r=sp.start;r<=sp.end;++r)				
				//if((sp.s.getNuclear(r).equals("n1")||sp.s.getNuclear(r).equals("n2"))&&(spansize<=5))
				//if((sp.s.getNuclear(r).equals("n1")||(sp.s.getNuclear(r).equals("pn")||sp.s.getNuclear(r).equals("n2"))))//&&(spansize<=5))
			 if(sp.s.getNuclear(r).equals("n2") )//&(spansize<=5))
					return (byte) TrueFalse.True.ordinal();
		
		
	}
	return (byte) TrueFalse.False.ordinal();
}*/
	
	
	

	@Override
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.NuclearAccent[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}
	
}
