package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.Nuclear;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_NuclearFirst extends AbstractPairEnumFeature<Nuclear> {
	private static final long serialVersionUID = 4988768071632378832L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_NuclearFirst(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"NuclearFirst", Nuclear.values());
		this.nodeExtractor=nodeExtractor;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab, List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.nuclearFirst!=null)
			return;
		inst.nuclearFirst=new byte[inst.nodes.length];
		Callable<Void> job=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int i=0;i<inst.nodes.length;++i)
					inst.nuclearFirst[i]=getByte(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(job));
	}

	private byte getByte(INode iNode) {
		Nuclear n = null;
		if(iNode.isVirtualNode())
			n=Nuclear.none;
		else {
			MNode mn=(MNode) iNode;
			Span sp=mn.span;
			int spansize=mn.span.size();
			
			for(int r=sp.start;r<=sp.end;++r)
				
				if (sp.s.getNuclear(r).equals("n2")&&(spansize<=5)){					//
					n=Nuclear.n2;
					return (byte) n.ordinal();
				}

				else if (sp.s.getNuclear(r).equals("n1")&&(spansize<=5)){
					n=Nuclear.n2; 	
					return (byte) n.ordinal();
				}
			
				else if (sp.s.getNuclear(r).equals("pn")&&(spansize<=5)){
					n=Nuclear.pn;
					return (byte) n.ordinal();
				}
			
				else if (sp.s.getNuclear(r).equals("0")){
					n=Nuclear.none;
				}
				else {n=Nuclear.none;}		
		}
		
		return (byte) n.ordinal();
	}

	@Override
//	int getIntValue(Instance inst,int headIdx,int depIdx,HOTState hotState) {
//		return inst.nuclear[nodeExtractor.getNodeInstIdx(headIdx,depIdx,hotState)];
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.nuclearFirst[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}
	
	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}

}