package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_ToBIFeature extends AbstractPairEnumFeature<TrueFalse>{
	
	private static final long serialVersionUID = 7460524611469166156L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_ToBIFeature(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"ToBIAccent", TrueFalse.values());
		this.nodeExtractor=nodeExtractor;
	}



	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}


	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab,List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.tobi!=null)
			return;
		Callable<Void> c=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				inst.tobi=new byte[inst.nodes.length];
				for(int i=0;i<inst.nodes.length;++i)
					inst.tobi[i]=ToBIAccent(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(c));
	}



	private static byte ToBIAccent(INode iNode) {
		final TrueFalse t;
	
		//version that works well! (based on syntactic head)
		if(iNode.isVirtualNode())
			t=TrueFalse.VNode;
		else {
			MNode mn=(MNode) iNode;
			int spansize=mn.span.size(); //uncomment this for previous version
			String hdAccent=mn.span.s.getAccent(mn.span.hd); 

			//if(hdAccent.equals("true"))
			if(hdAccent.equals("true")&&(spansize<=4))//remove second part for previous version
				t=TrueFalse.True;
			else
				t=TrueFalse.False;
		}
		return (byte) t.ordinal();
			
	}
			
			
			//version that did not perform so well, last accent in markable
		/*	t=TrueFalse.VNode;
			else {
				t=TrueFalse.False;
				MNode mn=(MNode) iNode;
				Span sp=mn.span;
				int spansize=mn.span.size(); //uncomment this for previous version
				String hdAccent=mn.span.s.getAccent(mn.span.hd); 
				for(int r=sp.start;r<=sp.end;++r){	
					
					String tobi=sp.s.getAccent(r);
					if(tobi.equals("true"))//&&(spansize<=4))
						t=TrueFalse.True;
					else
						t=TrueFalse.False;
				}
			}
				return (byte) t.ordinal();
				
*/
	



	@Override
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.tobi[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}
	
}

