

package ims.hotcoref.features;

import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

import ims.hotcoref.data.Instance;
import ims.hotcoref.data.Span;
import ims.hotcoref.decoder.HOTState;
import ims.hotcoref.features.enums.ToBI;
import ims.hotcoref.features.enums.TrueFalse;
import ims.hotcoref.features.extractors.PairTargetNodeExtractor;
import ims.hotcoref.mentiongraph.INode;
import ims.hotcoref.mentiongraph.MNode;
import ims.hotcoref.symbols.SymbolTable;
import ims.util.ThreadPoolSingleton;

public class F_ToBIType extends AbstractPairEnumFeature<ToBI> {
	private static final long serialVersionUID = 4988768071632378832L;

	private final PairTargetNodeExtractor nodeExtractor;
	
	protected F_ToBIType(PairTargetNodeExtractor nodeExtractor) {
		super(nodeExtractor.ts.toString()+"ToBiType", ToBI.values());
		this.nodeExtractor=nodeExtractor;
	}

	@Override
	public void XfillFillInstanceJobs(final Instance inst, SymbolTable symTab, List<Callable<Void>> l, List<Future<Void>> fjs) {
		if(inst.toBIType!=null)
			return;
		inst.toBIType=new byte[inst.nodes.length];
		Callable<Void> job=new Callable<Void>(){
			@Override
			public Void call() throws Exception {
				for(int i=0;i<inst.nodes.length;++i)
					inst.toBIType[i]=getByte(inst.nodes[i]);
				return null;
			}
		};
		fjs.add(ThreadPoolSingleton.getInstance().submit(job));
	}

	private byte getByte(INode iNode) {
		ToBI n = null;
		if(iNode.isVirtualNode())
			n=ToBI.NOTAPPLICABLE;
		else {
			MNode mn=(MNode) iNode;
			Span sp=mn.span;
			int spansize=mn.span.size();
		
			for(int r=sp.start;r<=sp.end;++r){	
				
				String tobi=sp.s.getTobiType(r);
				
				if (tobi.contains("||")){
					tobi="|"+sp.s.getTobiType(r).split("\\|\\|")[1];
				}
				
				if (tobi.equals("|H*|")){//&&spansize<=4)	{		//
					n=ToBI.HIGHLINK;
				}

				else if (tobi.equals("|L*|")){//&&spansize<=4){
					n=ToBI.LOWLINK;
				}
			
				else if (tobi.equals("|H*L|")){//&&spansize<=4){
					n=ToBI.FALL;
				}
			
				else if (tobi.equals("|L*H|")){//&&spansize<=4){
					n=ToBI.RISE;
				}
			
				else if (tobi.equals("|!H*L|")){//&&spansize<=4){
					n=ToBI.DOWNSTEPFALL;
				}
			
				else if (tobi.equals("|NONE|")){//&&spansize<=4){
					n=ToBI.NONE;
				}

				else if (tobi.equals("|L*HL|")){//&&spansize<=4){
					n=ToBI.LATEPEAK;
				}
				else if (tobi.equals("|HH*L|")){//&&spansize<=4){
					n=ToBI.EARLYPEAK;
				}
				else if (tobi.equals("N/A")){//&&spansize<=4){
					n=ToBI.NOTAPPLICABLE;
				}
		
				else {n=ToBI.NOTAPPLICABLE;}		
		}}
		
		return (byte) n.ordinal();
	}

	@Override
//	int getIntValue(Instance inst,int headIdx,int depIdx,HOTState hotState) {
//		return inst.tobi[nodeExtractor.getNodeInstIdx(headIdx,depIdx,hotState)];
	int getIntValue(Instance inst,int[] tnes,HOTState hotState) {
		return inst.toBIType[nodeExtractor.getNodeInstIdxPrecompArray(tnes)];
	}
	
	@Override
	public boolean firstOrderFeature() {
		return nodeExtractor.firstOrder();
	}

}