package ims.hotcoref.lang;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import ims.hotcoref.Options;
import ims.hotcoref.data.ISentence;
import ims.hotcoref.data.Span;
import ims.hotcoref.data.CFGTree.CFGNode;
import ims.hotcoref.data.CFGTree.NonTerminal;
import ims.hotcoref.data.CFGTree.Terminal;
import ims.hotcoref.features.FeatureSet;
import ims.hotcoref.features.enums.Gender;
import ims.hotcoref.features.enums.Num;
import ims.hotcoref.features.enums.SemanticClass;
import ims.hotcoref.headrules.HeadFinder;
import ims.hotcoref.headrules.HeadRules;
import ims.hotcoref.headrules.HeadRules.Direction;
import ims.hotcoref.headrules.HeadRules.Rule;
import ims.hotcoref.util.WordNetInterface;
import ims.hotcoref.util.BergsmaLinLookup;

public class German extends Language{
	private static final long serialVersionUID = 1L;

	private final BergsmaLinLookup lookup;
	
	public German(){
		try {
			lookup=new BergsmaLinLookup(Options.genderData);
		} catch (IOException e) {
			e.printStackTrace();
			throw new Error("!");
		}
	}
	@Override
	public FeatureSet getDefaultFeatureSet() {
		String[] names={
			"MFromHdLemmaMToEndsWithMatch",
			"MToMFromNumberMatch+MToMFromGenderMatch",
				"MFromHdLemmaMToExactStringMatch",
				"MToHdForm+MGPHdForm",
				"MToHdFormMFromSubStringMatch+MFromHdFormMToSubStringMatch+MToMTypeCoarse+MFromMTypeCoarse",
				"MToCFGPathToRootChoppedWD+MLinPreCFGPathToRootChoppedWD+MLinPreMToMentionDistBuckets+MToMTypeCoarse",
				"MGPHdForm+MFromHdForm",
				"MFromWholeSpanForm+MToWholeSpanForm   ",
				"MFromSPrPos+MToHdLemma",
				"MFromSPrPos",
				"MFromSPrForm+MToHdForm",
				"MFromSFoPos+MToPronounForm",
				"MFromSFForm+MToHdForm",
				"MFromParentNodeCFGSubCat+MToSPrPos",
				"MFromNodeCFGSubCat+MFromMToNested",
				"MFromMTypeCoarse+MFromHdForm+MToHdForm", 
				"MFromMToSameSpeaker+MFromPronounForm+MToPronounForm+MFromMToMentionDistBuckets",
				"MFromMToMentionDistBuckets+MToMTypeCoarse+MFromParentNodeCFGSubCat",
				"MFromMToMentionDistBuckets+MToMTypeCoarse",
				"MFromMToExactStringMatch+MFromMToDistanceBuckets",
				"MFromMToDistanceBuckets+MToHdForm",
				"MFromMToDistanceBuckets",
				"MFromMToCleverStringMatch+MToMTypeCoarse+MFromHdForm+MToHdForm",
				"MFromMToCleverStringMatch+MFromMTypeCoarse+MFromHdForm+MToHdForm",
				"MFromMToCFGSSPath",
				"MFromHdPos+MToHdPos",
				"MFromHdPos+MFromHdIPPos+MToHdBWUV",
				"MFromHdForm+MToHdForm",
				"MToNodeCFGSubCat+MFromMToNested",
				"MToParentNodeCFGSubCat+MFromMToMentionDistBuckets+MFromMTypeCoarse",
				"MToSFForm+MFromHdForm",

		};
		return FeatureSet.getFromNameArray(names);
	}

	
	@Override
	public boolean cleverStringMatch(Span ant,Span ana) {
		String s1=ana.getCleverString();
		String s2=ant.getCleverString();
		return s1.length()>0 && s1.equals(s2);
	}

	@Override
	public String computeCleverString(Span sp) {
		StringBuilder sb=new StringBuilder();
		for(int i=sp.start;i<=sp.end;++i){
			//XXX added "
			if(sp.s.getForm(i).equals("\"") || sp.s.getTag(i).equals("ART")  ||sp.s.getTag(i).equals(":") ||sp.s.getTag(i).equals(".")||sp.s.getTag(i).equals(","))
//			if(sp.s.tags[i].equals("DT") || sp.s.tags[i].equals("POS") ||sp.s.tags[i].equals(":") ||sp.s.tags[i].equals(".")||sp.s.tags[i].equals(","))
				continue;
			sb.append(sp.s.getForm(i)).append(" ");
		}
		return sb.toString().trim();
	}

		@Override
	public boolean isAlias(Span ant,Span ana) {
		return AliasStuff.isAlias(ant, ana);
	}


	@Override
	public void computeAtomicSpanFeatures(Span s) {
		s.isProperName=isProperName(s);
		s.isPronoun=isPronoun(s);
		s.isDefinite=isDefinite(s);
		s.isDemonstrative=isDemonstrative(s);
		s.isQuoted=isQuoted(s);
	}

	
	private boolean isQuoted(Span s) {
		boolean quoteBegin=false;
		boolean quoteEnd=false;
		for(int i=s.start-1;i>0;--i){
			if(s.s.getTag(i).equals("''"))
				return false;
			if(s.s.getTag(i).equals("``")||s.s.getTag(i).equals("\"")){
				quoteBegin=true;
				break;
			}
		}
		if(!quoteBegin)
			return false;
		for(int i=s.end+1;i<s.s.getLength();++i){
			if(s.s.getTag(i).equals("``"))
				return false;
			if(s.s.getTag(i).equals("''")||s.s.getTag(i).equals("\"")){
				quoteEnd=true;
				break;
			}
		}
		return quoteBegin && quoteEnd;
	}

	private SemanticClass pronounSemanticClassLookup(String lcSurfaceForm) {
		if(FEMALE_PRONOUNS_SET.contains(lcSurfaceForm))
			return SemanticClass.Female;
		if(MALE_PRONOUNS_SET.contains(lcSurfaceForm))
			return SemanticClass.Male;
		if (SINGULAR_PRONOUNS_SET.contains(lcSurfaceForm) && !lcSurfaceForm.startsWith("es"))
			return SemanticClass.Person;
		else
			return SemanticClass.Unknown;
	}

	@Override
	public Num lookupNumber(Span s) {
	
		if(s.isPronoun){
			String formLc=s.s.getForm(s.start).toLowerCase();
			if(SINGULAR_PRONOUNS_SET.contains(formLc))
				return Num.Sin;
			if(PLURAL_PRONOUNS_SET.contains(formLc))
				return Num.Plu;
		} 
		return Num.Unknown;
		
	}

	private static final Pattern MASC_TITLE_PATTERN=Pattern.compile("^(?:Hr\\.?|Herr)$");
	private static final Pattern FEM_TITLE_PATTERN=Pattern.compile("^(?:Fr\\.?|Frau)$");
	
	
/*	//old version
	@Override
	public Gender lookupGender(String str,Span s) {
		
		if (str.equals("fem"))
			return Gender.Fem;
		else if (str.equals("masc"))
				return Gender.Masc;
		else if (str.equals("neut"))
				return Gender.Neut;
		
		else if(s.isPronoun){
			String formLc=s.s.getForm(s.start).toLowerCase();
			if(MALE_PRONOUNS_SET.contains(formLc))
				return Gender.Masc;
			else if(FEMALE_PRONOUNS_SET.contains(formLc))
				return Gender.Fem;
			else if(NEUT_PRONOUNS_SET.contains(formLc))
				return Gender.Neut;
		}
		else if(s.isProperName){ //Might be a title of a person
			if(MASC_TITLE_PATTERN.matcher(s.s.getForm(s.start)).matches())
				return Gender.Masc;
			if(FEM_TITLE_PATTERN.matcher(s.s.getForm(s.start)).matches())
				return Gender.Fem;}
		return Gender.Unknown;	
	}*/

	//new version
	@Override
	public Gender lookupGender(Span s) {
		
		if(s.isPronoun){
			String formLc=s.s.getForm(s.start).toLowerCase();
			if(MALE_PRONOUNS_SET.contains(formLc))
				return Gender.Masc;
			else if(FEMALE_PRONOUNS_SET.contains(formLc))
				return Gender.Fem;
			else if(NEUT_PRONOUNS_SET.contains(formLc))
				return Gender.Neut;
		}
		else if(s.isProperName){ //Might be a title of a person
			if(MASC_TITLE_PATTERN.matcher(s.s.getForm(s.start)).matches())
				return Gender.Masc;
			if(FEM_TITLE_PATTERN.matcher(s.s.getForm(s.start)).matches())
				return Gender.Fem;}
		return Gender.Unknown;	
	}



	private static final Pattern DEMONSTRATIVE_PATTERN=Pattern.compile("^d(?:iese|ieses|dieser)$",Pattern.CASE_INSENSITIVE);
	private boolean isDemonstrative(Span s) {
		int len=s.end-s.start+1;
		if(len==1)
			return false;
		return DEMONSTRATIVE_PATTERN.matcher(s.s.getForm(s.start)).matches();
	}

	private boolean isDefinite(Span s) {
		int len=s.end-s.start+1;
		if(len==1)
			return false;
		return s.s.getForm(s.start).matches("^[Dd](er|ie|as)$");

	}

	private boolean isProperName(Span s) {
		int len=s.end-s.start+1;
		if(len>1){
			for(int i=s.start;i<s.end;++i){
				if(!s.s.getTag(i).startsWith("NE"))
					return true;
			}
			if( s.s.getTag(s.end).startsWith("NE"))
				return true;
			else
				return false;
		} else {
			return s.s.getTag(s.start).startsWith("NE");
		}
	}

	public boolean isPronoun(Span s) {
		
		if (s.size()>1)
			return false;
		
		else {
		//return s.s.getTag(s.hd).startsWith("PPER") ||s.s.getTag(s.hd).startsWith("PPOSAT")|| s.s.getTag(s.hd).startsWith("PRELS")||ALL_PRONOUNS.contains(s.s.getForm(s.hd).toLowerCase());
		return s.s.getTag(s.hd).equals("PPER")| s.s.getTag(s.hd).equals("PRELS")|| s.s.getTag(s.hd).equals("PPOSAT")||s.s.getTag(s.hd).equals("PRS")|| s.s.getTag(s.hd).equals("PWD")|| s.s.getTag(s.hd).equals("PWS")|| s.s.getTag(s.hd).equals("PRF")||ALL_PRONOUNS.contains(s.s.getForm(s.hd).toLowerCase());
		}
	}
	

		

	private static final String[] SINGULAR_PRONOUNS=new String[]{"ich","er","sie","es","mich","mir", "mein","meins","ihn","ihres","ihrs","seins","sein","ihr"};
	private static final String[] PLURAL_PRONOUNS=new String[]{"wir","unser","unsere","unsers","sie","ihnen","ihr","ihrs","uns"};
	private static final String[] MALE_PRONOUNS=new String[]{"er","ihn","sein","seins","seines"};
	private static final String[] FEMALE_PRONOUNS=new String[]{"sie","ihr","ihrs","ihres"};
	private static final String[] NEUT_PRONOUNS=new String[]{"es","sein","seins"};
	
	private static final Set<String> SINGULAR_PRONOUNS_SET=new HashSet<String>();
	private static final Set<String> PLURAL_PRONOUNS_SET=new HashSet<String>();
	private static final Set<String> MALE_PRONOUNS_SET=new HashSet<String>();
	private static final Set<String> FEMALE_PRONOUNS_SET=new HashSet<String>();
	private static final Set<String> NEUT_PRONOUNS_SET=new HashSet<String>();
	
	public static final Set<String> ALL_PRONOUNS=new HashSet<String>();
	static {
		Collections.addAll(SINGULAR_PRONOUNS_SET,SINGULAR_PRONOUNS);
		Collections.addAll(PLURAL_PRONOUNS_SET,PLURAL_PRONOUNS);
		Collections.addAll(FEMALE_PRONOUNS_SET,FEMALE_PRONOUNS);
		Collections.addAll(MALE_PRONOUNS_SET,MALE_PRONOUNS);
		Collections.addAll(NEUT_PRONOUNS_SET,NEUT_PRONOUNS);
		
		//All below
		Collections.addAll(ALL_PRONOUNS,SINGULAR_PRONOUNS);
		Collections.addAll(ALL_PRONOUNS,PLURAL_PRONOUNS);
		Collections.addAll(ALL_PRONOUNS,MALE_PRONOUNS);
		Collections.addAll(ALL_PRONOUNS,FEMALE_PRONOUNS);
		String[] additionalPronouns=new String[]{"du", "dein", "deins","deines"};
		Collections.addAll(ALL_PRONOUNS,additionalPronouns);


	}
	
	static class AliasStuff{
		
		public static boolean isAlias(Span ant,Span ana){
			String antSFWTP=toSurfaceFormWithoutTrailingPossesives(ant);
			String anaSFWTP=toSurfaceFormWithoutTrailingPossesives(ana);
//			return comparePerson(antSFWTP.split(" "),anaSFWTP.split(" ")) || compareOrg(antSFWTP,anaSFWTP);// || antSFWTP.equalsIgnoreCase(anaSFWTP);
			if(ant.ne==null || ana.ne==null || !ant.ne.getLabel().equals(ana.ne.getLabel()))
				return false;
			String neLbl=ant.ne.getLabel();	
			if(neLbl.equals("PER")){
				return comparePerson(antSFWTP.split(" "),anaSFWTP.split(" "));
			} else if(neLbl.equals("ORG")){
				return compareOrg(ant,ana);
			} else {
				return false;
			}
//				return toSurfaceFormWithoutTrailingPossesives(ant).equalsIgnoreCase(toSurfaceFormWithoutTrailingPossesives(ana));
//			}
		}
		
		private static boolean comparePerson(String[] ant, String[] ana) {
			return ant[ant.length-1].equals(ana[ana.length-1]);
		}

		private static boolean compareOrg(Span ant, Span ana) {
			String antStr=toSurfaceFormWithoutTrailingPossesives(ant);
			String anaStr=toSurfaceFormWithoutTrailingPossesives(ana);
			return compareOrg(antStr,anaStr);
		}
		
		private static boolean compareOrg(String antStr, String anaStr) {
//			String antStr=toSurfaceFormWithoutTrailingPossesives(ant);
//			String anaStr=toSurfaceFormWithoutTrailingPossesives(ana);
			if(antStr.replaceAll("\\.", "").equals(anaStr) ||
				anaStr.replaceAll("\\.", "").equals(antStr)){
				return true;
			} else {
				if(antStr.length()>anaStr.length()){
					String[] acr=getAcronyms(antStr);
					String s=loseInitialThe(anaStr);
					return matchesAny(s,acr);
				} else {
					String[] acr=getAcronyms(anaStr);
					String s=loseInitialThe(antStr);
					return matchesAny(s,acr);
				}
			}
		}
		

		public static String toSurfaceFormWithoutTrailingPossesives(Span s){
			StringBuilder sb=new StringBuilder();
			for(int i=s.start;i<=s.end;++i){
				if(s.s.getTag(i).equals("POS"))
					continue;
				sb.append(s.s.getForm(i)).append(" ");
			}
			return sb.toString().trim();
		}
		private static boolean matchesAny(String s,	String[] acronyms) {
			for(String acro:acronyms){
				if(s.equals(acro))
					return true;
			}
			return false;
		}
		
		private static String loseInitialThe(String s){
			String ret=s.replaceFirst("^[Dd](er|ie|das) ","");
			return ret;
		}

		private static String[] getAcronyms(String anaphorSurfaceForm) {
			String[] tokens=anaphorSurfaceForm.split(" ");
			StringBuilder a1=new StringBuilder();
			StringBuilder a2=new StringBuilder();
			StringBuilder a3=new StringBuilder();
			for(int i=0;i<tokens.length;++i){
				if(!tokens[i].toLowerCase().matches("(assoc|bros|co|coop|corp|devel|inc|llc|ltd)\\.?")){
					a1.append(tokens[i]);
					if(Character.isUpperCase(tokens[i].charAt(0))){
						a2.append(tokens[i].charAt(0));
						a3.append(tokens[i].charAt(0)).append(".");
					}
				}
			}
			return new String[]{ a1.toString(), a2.toString(), a3.toString()};
		}
	}



	@Override
	public int findNonTerminalHead(CFGNode n){
		return findEnglishCFGHead(n);
	}
	
	public static int findEnglishCFGHead(CFGNode n){
		if(n==null)
			return -1;
		if(n instanceof Terminal)
			return n.beg;
		NonTerminal nt=(NonTerminal) n;
		int h=headFinder.findHead(nt);
		if(h<1)
			return nt.end;
		else
			return h;
	}
	
	static final HeadFinder headFinder;
	static {
		Map<String,HeadRules> m=new HashMap<String,HeadRules>();
		String[] clearRules=new String[]{
		"PN	r	NE.*;NN.*;FM.*;TRUNC;APPR;APPRART;CARD;VAFIN;ADJA;ADJD;.*",
		"NM	r	NN.*;CARD;.*",
		"AVP	r	ADV;.*",
		"PP	r	APPRART;APPR;APPO;PROAV;NE;AOZR;PWAV;TRUNC;.*",
		"PP 	l	PROAV;.*",
		"NP	r	NN.*;NE.*;PPER;PIS;PDS;PRELS;PRF;PWS;PPOSS;FM;TRUNC;ADJA;CARD;PIAT;PWAV;PROAV;ADJD;ADV;APPRART;PDAT;.*",
		"CNP 	l	KON;.*",
		"MTA 	r	ADJA;.*",
		"CH	l	FM;NE;XY;CARD;.*",
		"AP	r	ADJA;NN.*;.*",
		//"EDITED	r	NN.*;NE.*;PPER;PIS;PDS;PRELS;PRF;PWS;PPOSS;FM;TRUNC;ADJA;CARD;PIAT;PWAV;PROAV;ADJD;ADV;APPRART;PDAT;.*",
		"S	r 	.*",
		"VP	r 	.*",
		"VROOT	r 	.*",
		"CVP	r 	.*",
		"CAP	r 	.*",
		"CS	r 	.*"};
		
		for(String line:clearRules){
			String[] a=line.split("\\t");
			String lbl=a[0];
			Direction d=(a[1].equals("r")?Direction.RightToleft:Direction.LeftToRight);
			String[] r=a[2].split(";");
			Rule[] rules=new Rule[r.length];
			int i=0;
			for(String s:r)
				rules[i++]=new Rule(d,Pattern.compile(s));
			m.put(lbl, new HeadRules(lbl,rules));
		}
		headFinder=new HeadFinder(m);
	}
	@Override
	public String getDefaultMarkableExtractors() {
		return "NT-NP,T-PPER,T-PPOSAT,T-PRELS,T-PDS,T-PRF,NT-PN";
	}
	
	public void preprocessSentence(ISentence s){
		if(s.getForm(1).equals("Mm"))
			s.setTag(1,"UH");
	}
	static final Set<String> nonReferentials=new HashSet<String>(Arrays.asList("es"));
	public Set<String> getNonReferentialTokenSet() {
		return nonReferentials;
	}
	
	
	

	@Override
	public String getDefaultEdgeCreators() {
		return "LeftGraph";
	}
	
	private static final Pattern COORD_TOKEN_PATTERN=Pattern.compile("(und|aber|oder|,)",Pattern.CASE_INSENSITIVE);
	public boolean isCoordToken(String string) {
		return COORD_TOKEN_PATTERN.matcher(string).matches();
	}
}